import re
from json import loads
from typing import Callable, Literal

from pydantic import BaseModel
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

_un_regex = re.compile(r"^[a-z_]([a-z0-9_-]{0,31}|[a-z0-9_-]{0,30}\$)$")
_snake_regex = re.compile(r"(?<!^)(?=[A-Z])")


def _is_octal(count: int) -> Callable[..., str]:
    def _check_is_octal(v: str) -> str:
        assert len(v) == count
        assert v[0] == "0"
        assert v
        for _ in range(count):
            assert 0 <= int(v[_]) <= 7
        return v

    return _check_is_octal


def _is_username(v: str) -> str:
    assert re.match(_un_regex, v)
    return v


def _is_username_or_usernamegroup(v: str) -> str:
    splits = v.split(":")
    assert len(splits) == 1 or len(splits) == 2
    assert re.match(_un_regex, splits[0])
    return v


def _is_host_port(v: str) -> str:
    splits = v.split(":")
    assert 0 < int(splits[1]) < 65_535
    return v


Octal = Annotated[str, AfterValidator(_is_octal(4))]
OctalUmask = Annotated[str, AfterValidator(_is_octal(3))]
UnixUserNameOrGroup = Annotated[str, AfterValidator(_is_username_or_usernamegroup)]
UnixUserName = Annotated[str, AfterValidator(_is_username)]
HostPort = Annotated[str, AfterValidator(_is_host_port)]
LogLevel = Literal["critical", "error", "warn", "info", "debug", "trace", "blather"]
Signal = Literal["TERM", "HUP", "INT", "QUIT", "KILL", "USR1", "USR2"]
EventType = Literal[
    "PROCESS_STATE",
    "PROCESS_STATE_STARTING",
    "PROCESS_STATE_RUNNING",
    "PROCESS_STATE_BACKOFF",
    "PROCESS_STATE_STOPPING",
    "PROCESS_STATE_EXITED",
    "PROCESS_STATE_STOPPED",
    "PROCESS_STATE_FATAL",
    "PROCESS_STATE_UNKNOWN",
    "REMOTE_COMMUNICATION",
    "PROCESS_LOG",
    "PROCESS_LOG_STDOUT",
    "PROCESS_LOG_STDERR",
    "PROCESS_COMMUNICATION",
    "PROCESS_COMMUNICATION_STDOUT",
    "PROCESS_COMMUNICATION_STDERR",
    "SUPERVISOR_STATE_CHANGE",
    "SUPERVISOR_STATE_CHANGE_RUNNING",
    "SUPERVISOR_STATE_CHANGE_STOPPING",
    "TICK",
    "TICK_5",
    "TICK_60",
    "TICK_3600",
    "PROCESS_GROUP",
    "PROCESS_GROUP_ADDED",
    "PROCESS_GROUP_REMOVED",
]
SupervisorLocation = Literal["local", "remote"]


class _BaseCfgModel(BaseModel):
    def to_cfg(self, key: str = "") -> str:
        ret = f"[{_snake_regex.sub('_', self.__class__.__name__.replace('Configuration', '')).lower()}{':' + key if key else ''}]"
        # round trip to json so we're fully
        # cfg-compatible
        for k, v in loads(self.model_dump_json()).items():
            if v is not None:
                if isinstance(v, bool):
                    ret += f"\n{k}={str(v).lower()}"
                else:
                    ret += f"\n{k}={v}"
        return ret.strip() + "\n"



from typing import List, Optional

from pydantic import Field, SecretStr, field_serializer, field_validator, model_validator

from .base import HostPort, Signal, SupervisorLocation, UnixUserName, _BaseCfgModel

__all__ = ("ConvenienceConfiguration",)


class ConvenienceConfiguration(_BaseCfgModel):
    """Convenience layer, settings that MUST be set when running via convenience API"""

    ############
    # programs #
    ############
    # autostart = False
    # autorestart = False
    startsecs: Optional[int] = Field(
        default=1,
        description="The total number of seconds which the program needs to stay running after a startup to consider the start successful (moving the process from the STARTING state to the RUNNING state). Set to 0 to indicate that the program needn’t stay running for any particular amount of time. Even if a process exits with an “expected” exit code (see exitcodes), the start will still be considered a failure if the process exits quicker than startsecs.",
    )
    startretries: Optional[int] = Field(
        default=None,
        description="The number of serial failure attempts that supervisord will allow when attempting to start the program before giving up and putting the process into an FATAL state. After each failed restart, process will be put in BACKOFF state and each retry attempt will take increasingly more time.",
    )
    exitcodes: Optional[List[int]] = Field(
        default=[0],
        description="The list of “expected” exit codes for this program used with autorestart. If the autorestart parameter is set to unexpected, and the process exits in any other way than as a result of a supervisor stop request, supervisord will restart the process if it exits with an exit code that is not defined in this list.",
    )
    stopsignal: Optional[Signal] = Field(
        default="TERM",
        description="The signal used to kill the program when a stop is requested. This can be specified using the signal’s name or its number. It is normally one of: TERM, HUP, INT, QUIT, KILL, USR1, or USR2.",
    )
    stopwaitsecs: Optional[int] = Field(
        default=30,
        description="The number of seconds to wait for the OS to return a SIGCHLD to supervisord after the program has been sent a stopsignal. If this number of seconds elapses before supervisord receives a SIGCHLD from the process, supervisord will attempt to kill it with a final SIGKILL.",
    )

    ####################
    # inet_http_server #
    ####################
    # port not optional
    port: HostPort = Field(
        default="*:9001",
        description="A TCP host:port value or (e.g. 127.0.0.1:9001) on which supervisor will listen for HTTP/XML-RPC requests. supervisorctl will use XML-RPC to communicate with supervisord over this port. To listen on all interfaces in the machine, use :9001 or *:9001. Please read the security warning above.",
    )
    username: Optional[UnixUserName] = Field(default=None, description="The username required for authentication to the HTTP/Unix Server.")
    password: Optional[SecretStr] = Field(
        default=None,
        description="The password required for authentication to the HTTP/Unix server. This can be a cleartext password, or can be specified as a SHA-1 hash if prefixed by the string {SHA}. For example, {SHA}82ab876d1387bfafe46cc1c8a2ef074eae50cb1d is the SHA-stored version of the password “thepassword”. Note that hashed password must be in hex format.",
    )

    #################
    # rpc_interface #
    ###################
    rpcinterface_factory: str = Field(
        default="supervisor.rpcinterface:make_main_rpcinterface",
        description="pkg_resources “entry point” dotted name to your RPC interface’s factory function.",
    )

    #########
    # Other #
    #########
    local_or_remote: Optional[SupervisorLocation] = Field(
        default="local",
        description="Location of supervisor, either local for same-machine or remote. If same-machine, communicates via Unix sockets by default, if remote, communicates via inet http server",
    )
    host: str = Field(
        default="localhost",
        description="Hostname of the supervisor host. Used by the XMLRPC client",
    )
    protocol: str = Field(
        default="http",
        description="Protocol of the supervisor XMLRPC HTTP API. Used by the XMLRPC client",
    )
    rpcpath: str = Field(
        default="/RPC2",
        description="Path for supervisor XMLRPC HTTP API. Used by the XMLRPC client",
    )

    @model_validator(mode="after")
    def validate_remote_accessible(self):
        if self.local_or_remote == "remote" and self.port.startswith(("localhost", "127.0.0.1")):
            raise ValueError("Supervisor binds only to loopback (localhost/127.0.0.1), but asked for remote")
        if self.local_or_remote == "remote" and self.host.startswith(("localhost", "127.0.0.1")):
            raise ValueError("Supervisor client expecting hostname, got localhost/127.0.0.1")
        return self

    @field_serializer("exitcodes", when_used="json")
    def _dump_exitcodes(self, v):
        if v:
            return ",".join(str(_) for _ in v)
        return None

    @field_validator("exitcodes", mode="before")
    @classmethod
    def _load_exitcodes(cls, v):
        if isinstance(v, str):
            v = v.split(",")
        if isinstance(v, list):
            return [int(_) for _ in v]
        return v



from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import ProgramConfiguration


def test_inst():
    with raises(ValidationError):
        ProgramConfiguration()
    ProgramConfiguration(command="echo 'test'")


def test_dump_exitcodes():
    # ensure that converting exit codes to string does not regress
    # pydantic_core._pydantic_core.PydanticSerializationError: Error serializing to JSON:
    # PydanticSerializationError: Error calling function `_dump_exitcodes`:
    # TypeError: sequence item 0: expected str instance, int found
    ProgramConfiguration(command="echo 'test'", exitcodes=[0]).model_dump_json()


def test_autorestart_options():
    ProgramConfiguration(command="echo 'test'", autorestart=True).model_dump_json()
    ProgramConfiguration(command="echo 'test'", autorestart=False).model_dump_json()
    ProgramConfiguration(command="echo 'test'", autorestart="unexpected").model_dump_json()
    with raises(ValidationError):
        ProgramConfiguration(command="echo 'test'", autorestart="other").model_dump_json()


def test_cfg():
    c = ProgramConfiguration(command="echo 'test'")
    assert c.to_cfg("name").strip() == "[program:name]\ncommand=echo 'test'"



from datetime import UTC, datetime
from pathlib import Path
from unittest.mock import patch

from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import (
    EventListenerConfiguration,
    FcgiProgramConfiguration,
    GroupConfiguration,
    IncludeConfiguration,
    InetHttpServerConfiguration,
    ProgramConfiguration,
    RpcInterfaceConfiguration,
    SupervisorConfiguration,
    SupervisorctlConfiguration,
    SupervisordConfiguration,
    UnixHttpServerConfiguration,
)


def test_inst():
    with raises(ValidationError):
        SupervisorConfiguration()
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.datetime") as p2,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        p2.now.return_value = datetime(2000, 1, 1, 0, 0, 0, 1, tzinfo=UTC)
        c = SupervisorConfiguration(program={"test": ProgramConfiguration(command="test")})
        assert str(c.working_dir) == str(pth / "supervisor-2000-01-01T00-00-00")
        assert str(c.config_path) == str(pth / "supervisor-2000-01-01T00-00-00" / "supervisor.cfg")


def test_cfg_roundtrip_json():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.datetime") as p2,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        p2.now.return_value = datetime(2000, 1, 1, 0, 0, 0, 1, tzinfo=UTC)
        c = SupervisorConfiguration(program={"test": ProgramConfiguration(command="test")})
        assert c.model_validate_json(c.model_dump_json()) == c


def test_cfg():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.datetime") as p2,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        p2.now.return_value = datetime(2000, 1, 1, 0, 0, 0, 1, tzinfo=UTC)
        c = SupervisorConfiguration(program={"test": ProgramConfiguration(command="test")})
        assert (
            c.to_cfg().strip()
            == """[supervisord]
logfile={dir}/supervisord.log
pidfile={dir}/supervisord.pid
directory={dir}

[supervisorctl]

[program:test]
command=test
directory={dir}/test""".format(dir=str(pth / "supervisor-2000-01-01T00-00-00"))
        )


def test_cfg_all():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.datetime") as p2,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        p2.now.return_value = datetime(2000, 1, 1, 0, 0, 0, 1, tzinfo=UTC)
        c = SupervisorConfiguration(
            unix_http_server=UnixHttpServerConfiguration(
                file="/a/test/file",
                chmod="0777",
                chown="test",
                username="test",
                password="testpw",
            ),
            inet_http_server=InetHttpServerConfiguration(port="127.0.0.1:8000", username="test", password="testpw"),
            supervisord=SupervisordConfiguration(),
            supervisorctl=SupervisorctlConfiguration(username="test", password="testpw"),
            include=IncludeConfiguration(files=["a/test/file", "another/test/file"]),
            program={"test": ProgramConfiguration(command="test")},
            group={"testgroup": GroupConfiguration(programs=["test"])},
            fcgiprogram={"testfcgi": FcgiProgramConfiguration(command="echo 'test'", socket="test")},
            eventlistener={"testeventlistener": EventListenerConfiguration(command="echo 'test'")},
            rpcinterface={"testrpcinterface": RpcInterfaceConfiguration(rpcinterface_factory="a.test.module")},
        )
        print(c.to_cfg().strip())
        assert (
            c.to_cfg().strip()
            == """[unix_http_server]
file=/a/test/file
chmod=0777
chown=test
username=test
password=testpw

[inet_http_server]
port=127.0.0.1:8000
username=test
password=testpw

[supervisord]
logfile={dir}/supervisord.log
pidfile={dir}/supervisord.pid
directory={dir}

[supervisorctl]
username=test
password=testpw

[include]
files=a/test/file another/test/file

[program:test]
command=test
directory={dir}/test

[group:testgroup]
programs=test

[fcgi-program:testfcgi]
command=echo 'test'
socket=test

[eventlistener:testeventlistener]
command=echo 'test'

[rpcinterface:testrpcinterface]
supervisor.rpcinterface_factory=a.test.module""".format(dir=str(pth / "supervisor-2000-01-01T00-00-00"))
        )



from pathlib import Path
from typing import Optional

from pydantic import AnyUrl, Field, SecretStr, field_serializer

from .base import UnixUserName, _BaseCfgModel

__all__ = ("SupervisorctlConfiguration",)


class SupervisorctlConfiguration(_BaseCfgModel):
    serverurl: Optional[AnyUrl] = Field(
        default=None,
        description="The URL that should be used to access the supervisord server, e.g. http://localhost:9001. For UNIX domain sockets, use unix:///absolute/path/to/file.sock.",
    )
    username: Optional[UnixUserName] = Field(
        default=None,
        description="The username to pass to the supervisord server for use in authentication. This should be same as username from the supervisord server configuration for the port or UNIX domain socket you’re attempting to access.",
    )
    password: Optional[SecretStr] = Field(
        default=None,
        description="The password to pass to the supervisord server for use in authentication. This should be the cleartext version of password from the supervisord server configuration for the port or UNIX domain socket you’re attempting to access. This value cannot be passed as a SHA hash. Unlike other passwords specified in this file, it must be provided in cleartext.",
    )
    prompt: Optional[str] = Field(default=None, description="String used as supervisorctl prompt.")
    history_file: Optional[Path] = Field(
        default=None,
        description="A path to use as the readline persistent history file. If you enable this feature by choosing a path, your supervisorctl commands will be kept in the file, and you can use readline (e.g. arrow-up) to invoke commands you performed in your last supervisorctl session.",
    )

    @field_serializer("password", when_used="json")
    def _dump_password(self, v):
        return v.get_secret_value() if v else None



from logging import getLogger
from pathlib import Path
from time import sleep
from typing import Callable, Optional, Union

from typer import Argument, Exit, Option, Typer
from typing_extensions import Annotated

from ..client import SupervisorRemoteXMLRPCClient
from ..config import SupervisorConvenienceConfiguration
from .common import SupervisorTaskStep

log = getLogger(__name__)

__all__ = (
    "write_supervisor_config",
    "start_supervisor",
    "start_programs",
    "check_programs",
    "stop_programs",
    "restart_programs",
    "stop_supervisor",
    "kill_supervisor",
    "remove_supervisor_config",
    "main",
)


def _check_exists(cfg: SupervisorConvenienceConfiguration) -> bool:
    if cfg.working_dir.exists() and cfg.config_path.exists():
        # its probably already been written
        return True
    return False


def _check_same(cfg: SupervisorConvenienceConfiguration) -> bool:
    if _check_exists(cfg) and cfg.config_path.read_text().strip() == cfg.to_cfg().strip():
        # same file contents
        return True
    elif not _check_exists(cfg):
        return True
    return False


def _check_running(cfg: SupervisorConvenienceConfiguration) -> bool:
    if _check_same(cfg):
        return cfg.running()
    return False


def _wait_or_while(until: Callable, unless: Optional[Callable] = None, timeout: int = 5) -> bool:
    for _ in range(timeout):
        if until():
            return True
        if unless and unless():
            return False
        sleep(1)
    return False


def _raise_or_exit(val: bool, exit: bool):
    if exit:
        raise Exit(int(not val))
    return val


def _load_or_pass(cfg: Union[str, SupervisorConvenienceConfiguration]) -> SupervisorConvenienceConfiguration:
    if isinstance(cfg, Path):
        cfg = SupervisorConvenienceConfiguration.model_validate_json(cfg.read_text())
    if isinstance(cfg, str):
        cfg = SupervisorConvenienceConfiguration.model_validate_json(cfg)
    if not isinstance(cfg, SupervisorConvenienceConfiguration):
        raise NotImplementedError
    return cfg


def write_supervisor_config(cfg_json: str, _exit: Annotated[bool, Argument(hidden=True)] = True):
    """Write a SupervisorConvenienceConfiguration JSON as a supervisor config file

    Args:
        cfg_json (str): JSON string of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg_json)
    if not _check_same(cfg_obj):
        log.critical("Configs don't match")
    cfg_obj._write_self()
    return _raise_or_exit(True, _exit)


def start_supervisor(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Start a supervisor instance using supervisord in background

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    if not _check_same(cfg_obj):
        log.critical("Configs don't match")
    if _check_running(cfg_obj):
        return _raise_or_exit(True, _exit)
    cfg_obj.start(daemon=True)
    running = _wait_or_while(until=lambda: cfg_obj.running(), timeout=30)
    if not running:
        log.critical("Still not running 30s after start command!")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def start_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Start all programs in the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)

    ret = client.startAllProcesses()
    log.info(ret)

    _wait_or_while(
        until=lambda: all(_.running() for _ in client.getAllProcessInfo()),
        unless=lambda: any(_.stopped() for _ in client.getAllProcessInfo()),
        timeout=5,
    )
    all_ok = _wait_or_while(
        until=lambda: all(_.ok(ok_exitstatuses=cfg_obj.airflow.exitcodes) for _ in client.getAllProcessInfo()),
        unless=lambda: any(_.bad(ok_exitstatuses=cfg_obj.airflow.exitcodes) for _ in client.getAllProcessInfo()),
        timeout=10,
    )
    if not all_ok:
        for r in client.getAllProcessInfo():
            log.info(r.model_dump_json())
        log.warning("not all processes started")
        return _raise_or_exit(False, _exit)
    log.info("all processes started")
    return _raise_or_exit(True, _exit)


def check_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    check_running: bool = False,
    check_done: bool = False,
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Check if programs are in a good state.

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
        check_running (bool, optional): if true, only return true if they're running
        check_done (bool, optional): if true, only return true if they're done (cleanly)
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)

    ret = client.getAllProcessInfo()
    for r in ret:
        log.info(r.model_dump_json())

    ok = False
    if check_running:
        if all(p.running() for p in ret):
            log.info("all processes running")
            ok = True
        else:
            log.warning("not all processes running")
    elif check_done:
        if all(p.done(ok_exitstatuses=cfg_obj.airflow.exitcodes) for p in ret):
            log.info("all processes done")
            ok = True
        else:
            log.info("not all processes done")
    else:
        if all(p.ok(ok_exitstatuses=cfg_obj.airflow.exitcodes) for p in ret):
            log.info("all processes ok")
            ok = True
        else:
            log.info("not all processes ok")
    return _raise_or_exit(ok, _exit)


def stop_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Stop all programs in the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    client = SupervisorRemoteXMLRPCClient(cfg=cfg_obj)

    ret = client.stopAllProcesses()
    log.info(ret)

    all_stopped = _wait_or_while(until=lambda: all(_.stopped() for _ in client.getAllProcessInfo()), timeout=10)
    if not all_stopped:
        for r in client.getAllProcessInfo():
            log.info(r.model_dump_json())
        log.warning("not all processes stopped")
        return _raise_or_exit(False, _exit)
    log.info("all processes stopped")
    return _raise_or_exit(True, _exit)


def restart_programs(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    force: bool = False,
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Restart all programs in the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
        force (bool, optional): if true, force restart. Defaults to False.
    """
    if force:
        if not stop_programs(cfg, False):
            log.warning("could not stop programs")
            return _raise_or_exit(False, _exit)
    if not start_programs(cfg, False):
        log.warning("could not start programs")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def stop_supervisor(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Stop the supervisor instance

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    cfg_obj.stop()
    not_running = _wait_or_while(until=lambda: not cfg_obj.running(), timeout=30)
    if not not_running:
        log.critical("Still running 30s after stop command!")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def kill_supervisor(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Kill the supervisor instance with os.kill

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    try:
        if not stop_programs(cfg, False):
            log.warning("could not stop programs")
    except ConnectionRefusedError:
        # supervisor already down, continue
        ...

    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    cfg_obj.kill()
    still_running = _wait_or_while(until=lambda: not cfg_obj.running(), timeout=30)
    if still_running:
        log.critical("Still running 30s after kill command!")
        return _raise_or_exit(False, _exit)
    return _raise_or_exit(True, _exit)


def remove_supervisor_config(
    cfg: Annotated[Path, Option(exists=True, file_okay=True, dir_okay=False, writable=False, readable=True, resolve_path=True)] = Path(
        "pydantic.json"
    ),
    _exit: Annotated[bool, Argument(hidden=True)] = True,
):
    """Remove the supervisor config file and working directory

    Args:
        cfg (Annotated[Path, Option, optional): Path to JSON file of SupervisorConvenienceConfiguration
    """
    # NOTE: typer does not support union types
    cfg_obj = _load_or_pass(cfg)
    still_running = stop_supervisor(cfg_obj, _exit=False)
    if still_running:
        still_running = kill_supervisor(cfg_obj, _exit=False)

    if still_running:
        return _raise_or_exit(False, _exit)

    # TODO move to config
    sleep(5)

    # TODO make optional
    cfg_obj.rmdir()
    return _raise_or_exit(True, _exit)


def _add_to_typer(app, command: SupervisorTaskStep, foo):
    """Helper function to ensure correct command names"""
    app.command(command)(foo)


def main():
    app = Typer()
    _add_to_typer(app, "configure-supervisor", write_supervisor_config)
    _add_to_typer(app, "start-supervisor", start_supervisor)
    _add_to_typer(app, "start-programs", start_programs)
    _add_to_typer(app, "stop-programs", stop_programs)
    _add_to_typer(app, "check-programs", check_programs)
    _add_to_typer(app, "restart-programs", restart_programs)
    _add_to_typer(app, "stop-supervisor", stop_supervisor)
    _add_to_typer(app, "force-kill", kill_supervisor)
    _add_to_typer(app, "unconfigure-supervisor", remove_supervisor_config)
    app()



from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import GroupConfiguration


def test_inst():
    with raises(ValidationError):
        GroupConfiguration()
    GroupConfiguration(programs=["test"])


def test_cfg():
    c = GroupConfiguration(programs=["test"])
    assert c.to_cfg("name").strip() == "[group:name]\nprograms=test"



import socket
from tempfile import NamedTemporaryFile
from time import sleep
from typing import Iterator

from pytest import fixture

from supervisor_pydantic import ConvenienceConfiguration, ProgramConfiguration, SupervisorConvenienceConfiguration


@fixture(scope="module")
def open_port() -> int:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


@fixture(scope="module")
def permissioned_open_port() -> int:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


@fixture(scope="module")
def supervisor_convenience_configuration(open_port: int) -> Iterator[SupervisorConvenienceConfiguration]:
    with NamedTemporaryFile("w", suffix=".cfg") as tf:
        cfg = SupervisorConvenienceConfiguration(
            airflow=ConvenienceConfiguration(port=f"*:{open_port}"),
            path=tf.name,
            program={
                "test": ProgramConfiguration(
                    command="bash -c 'sleep 1; exit 1'",
                )
            },
        )
        yield cfg


@fixture(scope="module")
def permissioned_supervisor_convenience_configuration(
    permissioned_open_port: int,
) -> Iterator[SupervisorConvenienceConfiguration]:
    with NamedTemporaryFile("w", suffix=".cfg") as tf:
        cfg = SupervisorConvenienceConfiguration(
            airflow=ConvenienceConfiguration(port=f"*:{permissioned_open_port}", username="user1", password="testpassword1"),
            path=tf.name,
            program={
                "test": ProgramConfiguration(
                    command="bash -c 'sleep 1; exit 1'",
                )
            },
        )
        yield cfg


@fixture(scope="module")
def supervisor_instance(
    supervisor_convenience_configuration: SupervisorConvenienceConfiguration,
) -> Iterator[SupervisorConvenienceConfiguration]:
    cfg = supervisor_convenience_configuration
    cfg.write()
    cfg.start(daemon=False)
    for _ in range(5):
        if not cfg.running():
            sleep(1)
    yield cfg
    cfg.kill()


@fixture(scope="module")
def permissioned_supervisor_instance(
    permissioned_supervisor_convenience_configuration: SupervisorConvenienceConfiguration,
) -> Iterator[SupervisorConvenienceConfiguration]:
    cfg = permissioned_supervisor_convenience_configuration
    cfg.write()
    cfg.start(daemon=False)
    for _ in range(5):
        if not cfg.running():
            sleep(1)
    yield cfg
    cfg.kill()



from supervisor_pydantic import InetHttpServerConfiguration


def test_inst():
    InetHttpServerConfiguration()
    InetHttpServerConfiguration(port="127.0.0.1:8000")
    InetHttpServerConfiguration(port="127.0.0.1:8000", username="test")
    InetHttpServerConfiguration(port="127.0.0.1:8000", username="test", password="testpw")


def test_cfg():
    c = InetHttpServerConfiguration()
    assert c.to_cfg().strip() == """[inet_http_server]"""

    c = InetHttpServerConfiguration(port="127.0.0.1:8000", username="test", password="testpw")
    assert (
        c.to_cfg()
        == """[inet_http_server]
port=127.0.0.1:8000
username=test
password=testpw
"""
    )



import os
from datetime import datetime, time, timedelta, timezone
from logging import getLogger
from pathlib import Path
from shutil import rmtree
from signal import SIGKILL, SIGTERM
from subprocess import Popen
from tempfile import gettempdir
from typing import Dict, Optional

from hydra import compose, initialize_config_dir
from hydra.utils import instantiate
from pydantic import BaseModel, Field, PrivateAttr, model_validator

from ..exceptions import ConfigNotFoundError
from ..utils import _get_calling_file
from .convenience import ConvenienceConfiguration
from .eventlistener import EventListenerConfiguration
from .fcgiprogram import FcgiProgramConfiguration
from .group import GroupConfiguration
from .include import IncludeConfiguration
from .inet_http_server import InetHttpServerConfiguration
from .program import ProgramConfiguration
from .rpcinterface import RpcInterfaceConfiguration
from .supervisorctl import SupervisorctlConfiguration
from .supervisord import SupervisordConfiguration
from .unix_http_server import UnixHttpServerConfiguration

__all__ = (
    "SupervisorConfiguration",
    "SupervisorConvenienceConfiguration",
    "load_config",
    "load_convenience_config",
)

_log = getLogger(__name__)


class SupervisorConfiguration(BaseModel):
    def to_cfg(self) -> str:
        ret = ""
        if self.unix_http_server:
            ret += self.unix_http_server.to_cfg() + "\n"
        if self.inet_http_server:
            ret += self.inet_http_server.to_cfg() + "\n"
        if self.supervisord:
            ret += self.supervisord.to_cfg() + "\n"
        if self.supervisorctl:
            ret += self.supervisorctl.to_cfg() + "\n"
        if self.include:
            ret += self.include.to_cfg() + "\n"
        for k, v in self.program.items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.group or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.fcgiprogram or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.eventlistener or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.rpcinterface or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        return ret

    # supervisor setup
    unix_http_server: Optional[UnixHttpServerConfiguration] = Field(default=None)
    inet_http_server: Optional[InetHttpServerConfiguration] = Field(default=None)
    supervisord: SupervisordConfiguration = Field(default=SupervisordConfiguration())
    supervisorctl: SupervisorctlConfiguration = Field(default=SupervisorctlConfiguration())
    include: Optional[IncludeConfiguration] = Field(default=None)

    program: Dict[str, ProgramConfiguration]
    group: Optional[Dict[str, GroupConfiguration]] = Field(default=None)

    fcgiprogram: Optional[Dict[str, FcgiProgramConfiguration]] = Field(default=None)
    eventlistener: Optional[Dict[str, EventListenerConfiguration]] = Field(default=None)
    rpcinterface: Optional[Dict[str, RpcInterfaceConfiguration]] = Field(default=None)

    # other configuration
    config_path: Optional[Path] = Field(default="", description="Path to supervisor configuration file")
    working_dir: Optional[Path] = Field(default="", description="Path to supervisor working directory")

    @model_validator(mode="after")
    def _setup_config_and_working_dir(self):
        now = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H-%M-%S")
        tempdir = Path(gettempdir()).resolve()

        if self.working_dir == "":
            if self.supervisord.directory:
                # use this as the dir
                self.working_dir = self.supervisord.directory
            else:
                self.working_dir = tempdir / f"supervisor-{now}"
                self.supervisord.directory = self.working_dir
            using_default_working_dir = True
        else:
            using_default_working_dir = False

        # force pidfile to be in working dir if not otherwise set
        if not self.supervisord.pidfile:
            self.supervisord.pidfile = self.working_dir / "supervisord.pid"

        # force logfile to be in working dir if not otherwise set
        if not self.supervisord.logfile:
            self.supervisord.logfile = self.working_dir / "supervisord.log"

        if self.config_path == "":
            if using_default_working_dir:
                self.config_path = self.working_dir / "supervisor.cfg"
            else:
                self.config_path = self.working_dir / f"supervisor-{now}.cfg"

        for name, program_config in self.program.items():
            if program_config.directory is None:
                program_config.directory = self.working_dir / name
        return self

    @classmethod
    def _find_parent_config_folder(cls, config_dir: str = "config", config_name: str = "", *, basepath: str = "", _offset: int = 2):
        if basepath:
            if basepath.endswith((".py", ".cfg", ".yml", ".yaml")):
                calling_dag = Path(basepath)
            else:
                calling_dag = Path(basepath) / "dummy.py"
        else:
            calling_dag = Path(_get_calling_file(offset=_offset))
        folder = calling_dag.parent.resolve()
        exists = (
            (folder / config_dir).exists()
            if not config_name
            else ((folder / config_dir / f"{config_name}.yml").exists() or (folder / config_dir / f"{config_name}.yaml").exists())
        )
        while not exists:
            folder = folder.parent
            if str(folder) == os.path.abspath(os.sep):
                raise ConfigNotFoundError(config_dir=config_dir, dagfile=calling_dag)
            exists = (
                (folder / config_dir).exists()
                if not config_name
                else ((folder / config_dir / f"{config_name}.yml").exists() or (folder / config_dir / f"{config_name}.yaml").exists())
            )

        config_dir = (folder / config_dir).resolve()
        if not config_name:
            return folder.resolve(), config_dir, ""
        elif (folder / config_dir / f"{config_name}.yml").exists():
            return folder.resolve(), config_dir, (folder / config_dir / f"{config_name}.yml").resolve()
        return folder.resolve(), config_dir, (folder / config_dir / f"{config_name}.yaml").resolve()

    @classmethod
    def load(
        cls: "SupervisorConfiguration",
        config_dir: str = "config",
        config_name: str = "",
        overrides: Optional[list[str]] = None,
        *,
        basepath: str = "",
        _offset: int = 3,
    ) -> "SupervisorConfiguration":
        overrides = overrides or []

        with initialize_config_dir(config_dir=str(Path(__file__).resolve().parent / "hydra"), version_base=None):
            if config_dir:
                hydra_folder, config_dir, _ = cls._find_parent_config_folder(
                    config_dir=config_dir, config_name=config_name, basepath=basepath, _offset=_offset
                )

                cfg = compose(config_name="base", overrides=[], return_hydra_config=True)
                searchpaths = cfg["hydra"]["searchpath"]
                searchpaths.extend([hydra_folder, config_dir])
                if config_name:
                    overrides = [
                        f"+config={config_name}",
                        *overrides.copy(),
                        f"hydra.searchpath=[{','.join(searchpaths)}]",
                    ]
                else:
                    overrides = [*overrides.copy(), f"hydra.searchpath=[{','.join(searchpaths)}]"]

            cfg = compose(config_name="base", overrides=overrides)
            config = instantiate(cfg)

            if not isinstance(config, cls):
                if issubclass(cls, type(config)):
                    config = config.model_dump()
                config = cls(**config)
            return config

    def write(self):
        _log.info(f"Making working dir: {self.working_dir}")
        self.working_dir.mkdir(parents=True, exist_ok=True)
        for program_name, program_config in self.program.items():
            if program_config.directory:
                _log.info(f"Making program dir ({program_name}): {program_config.directory}")
                program_config.directory.mkdir(exist_ok=True)
        _log.info(f"Writing config file: {self.config_path}")
        self.config_path.write_text(self.to_cfg())

    def rmdir(self):
        if not self.running():
            _log.info(f"Removing working dir: {self.working_dir}")
            rmtree(self.working_dir)

    def start(self, daemon: bool = False):
        _log.info(f"Starting supervisord: {self.config_path}")
        if not self.running():
            if daemon is False:
                Popen(f"supervisord -n -c {str(self.config_path)}", shell=True)
                return
            Popen(f"supervisord -c {str(self.config_path)}", close_fds=True, shell=True)

    def running(self):
        # grab the pidfile, find the process with the pid, and kill
        _log.info(f"Checking supervisord: {self.supervisord.pidfile}")
        if not self.supervisord.pidfile.exists():
            return False
        try:
            os.kill(int(self.supervisord.pidfile.read_text()), 0)
        except OSError:
            return False
        return True

    def stop(self):
        if self.running():
            # grab the pidfile, find the process with the pid, and kill with SIGTERM
            _log.info(f"Stopping supervisord: {self.supervisord.pidfile}")
            os.kill(int(self.supervisord.pidfile.read_text()), SIGTERM)

    def kill(self):
        if self.running():
            # grab the pidfile, find the process with the pid, and kill with SIGKILL
            _log.info(f"Killing supervisord: {self.supervisord.pidfile}")
            os.kill(int(self.supervisord.pidfile.read_text()), SIGKILL)


class SupervisorConvenienceConfiguration(SupervisorConfiguration):
    airflow: ConvenienceConfiguration = Field(
        default_factory=ConvenienceConfiguration, description="Required configurations for convenience integration"
    )

    # Passthrough to PythonSensor in airflow-ha
    check_interval: timedelta = Field(default=timedelta(seconds=5), description="Interval between supervisor program status checks")
    check_timeout: timedelta = Field(default=timedelta(hours=8), description="Timeout to wait for supervisor program status checks")

    # HighAvailabilityOperator custom args
    runtime: Optional[timedelta] = Field(default=None, description="Max runtime of Supervisor job")
    endtime: Optional[time] = Field(default=None, description="End time of Supervisor job")
    maxretrigger: Optional[int] = Field(
        default=None,
        description="Max number of retriggers of Supervisor job (e.g. max number of checks separated by `check_interval`)",
    )

    _pydantic_path: Path = PrivateAttr(default="pydantic.json")

    def _write_self(self):
        # TODO make config driven
        self.write()
        _log.info(f"Writing model json: {self._pydantic_path}")
        Path(self._pydantic_path).write_text(self.model_dump_json())

    @model_validator(mode="after")
    def _setup_airflow_defaults(self):
        """Method to overload configuration with values needed for the setup
        of airflow tasks that we construct"""
        # inet_http_server
        if not self.inet_http_server:
            self.inet_http_server = InetHttpServerConfiguration()

        self.inet_http_server.port = self.airflow.port
        self.inet_http_server.username = self.airflow.username
        self.inet_http_server.password = self.airflow.password

        self.supervisorctl.serverurl = f"{self.airflow.protocol}://{self.airflow.host}:{self.airflow.port.split(':')[-1]}/"

        # rpcinterface
        if not self.rpcinterface:
            self.rpcinterface = {"supervisor": RpcInterfaceConfiguration()}
        self.rpcinterface["supervisor"].rpcinterface_factory = self.airflow.rpcinterface_factory

        # supervisord
        self.supervisord.nodaemon = False
        self.supervisord.identifier = "supervisor"

        # programs
        for config in self.program.values():
            config.autostart = False
            config.autorestart = False
            config.startsecs = self.airflow.startsecs
            config.startretries = self.airflow.startretries
            config.exitcodes = self.airflow.exitcodes
            config.stopsignal = self.airflow.stopsignal
            config.stopwaitsecs = self.airflow.stopwaitsecs

        # other
        if str(self.working_dir) not in str(self._pydantic_path):
            self._pydantic_path = self.working_dir / "pydantic.json"
        return self


load_config = SupervisorConfiguration.load
load_convenience_config = SupervisorConvenienceConfiguration.load



from .base import HostPort, LogLevel, Octal, OctalUmask, Signal, SupervisorLocation, UnixUserName, UnixUserNameOrGroup
from .convenience import ConvenienceConfiguration
from .eventlistener import EventListenerConfiguration
from .fcgiprogram import FcgiProgramConfiguration
from .group import GroupConfiguration
from .include import IncludeConfiguration
from .inet_http_server import InetHttpServerConfiguration
from .program import ProgramConfiguration
from .rpcinterface import RpcInterfaceConfiguration
from .supervisor import (
    SupervisorConfiguration,
    SupervisorConvenienceConfiguration,
    load_config,
    load_convenience_config,
)
from .supervisorctl import SupervisorctlConfiguration
from .supervisord import SupervisordConfiguration
from .unix_http_server import UnixHttpServerConfiguration



from subprocess import check_call

from supervisor_pydantic.config import SupervisorConvenienceConfiguration
from supervisor_pydantic.convenience.commands import start_supervisor, stop_supervisor, write_supervisor_config


def test_command():
    assert check_call(["_supervisor_convenience", "--help"]) == 0


def test_write(supervisor_convenience_configuration: SupervisorConvenienceConfiguration):
    json = supervisor_convenience_configuration.model_dump_json()
    assert write_supervisor_config(json, _exit=False)
    assert supervisor_convenience_configuration._pydantic_path.read_text().strip() == json
    supervisor_convenience_configuration.rmdir()


def test_start_stop(supervisor_convenience_configuration: SupervisorConvenienceConfiguration):
    json = supervisor_convenience_configuration.model_dump_json()
    assert write_supervisor_config(json, _exit=False)
    assert supervisor_convenience_configuration._pydantic_path.read_text().strip() == json
    assert start_supervisor(supervisor_convenience_configuration._pydantic_path, _exit=False)
    assert stop_supervisor(supervisor_convenience_configuration._pydantic_path, _exit=False)
    supervisor_convenience_configuration.rmdir()



from pytest import raises

from supervisor_pydantic.config.base import _is_octal, _is_username, _is_username_or_usernamegroup


def test_is_octal():
    is_octal = _is_octal(4)
    assert is_octal("0700") == "0700"
    with raises(AssertionError):
        assert is_octal("0800")
    with raises(AssertionError):
        assert is_octal("1800")
    with raises(AssertionError):
        assert is_octal("000")


def test_is_username():
    assert _is_username("test")
    with raises(AssertionError):
        assert _is_username("test@123")
    with raises(AssertionError):
        assert _is_username("test:hey")


def test_is_username_or_usernamegroup():
    assert _is_username_or_usernamegroup("test")
    assert _is_username_or_usernamegroup("test:hey")
    with raises(AssertionError):
        assert _is_username_or_usernamegroup("test@123")



from inspect import currentframe


def _get_calling_file(offset: int = 2) -> str:
    cur_frame = currentframe()
    for _ in range(offset):
        if hasattr(cur_frame, "f_back") and cur_frame.f_back and hasattr(cur_frame.f_back, "f_globals"):
            cur_frame = cur_frame.f_back
        else:
            break
    return cur_frame.f_globals["__file__"]



__all__ = ("ConfigNotFoundError",)


class ConfigNotFoundError(RuntimeError):
    def __init__(self, config_dir, dagfile):
        super().__init__(f"Could not find `{config_dir}` directory on or above DAG: {dagfile}")



from typing import Any, Dict, Optional

from pydantic import Field

from .base import _BaseCfgModel

__all__ = ("RpcInterfaceConfiguration",)


class RpcInterfaceConfiguration(_BaseCfgModel):
    def to_cfg(self, key: str) -> str:
        # Overload to require key
        return super().to_cfg(key=key).replace("[rpc_interface", "[rpcinterface").replace("rpcinterface_factory=", "supervisor.rpcinterface_factory=")

    rpcinterface_factory: str = Field(
        default="supervisor.rpcinterface:make_main_rpcinterface",
        description="pkg_resources “entry point” dotted name to your RPC interface’s factory function.",
    )
    kwargs: Optional[Dict[str, Any]] = Field(default=None)  # TODO



from pydantic import ValidationError
from pytest import raises

from supervisor_pydantic import IncludeConfiguration


def test_inst():
    with raises(ValidationError):
        IncludeConfiguration()
    IncludeConfiguration(files=[""])


def test_cfg():
    c = IncludeConfiguration(files=["a/test/file", "another/test/file"])
    assert c.to_cfg().strip() == "[include]\nfiles=a/test/file another/test/file"



import xmlrpc
from time import sleep

import pytest

from supervisor_pydantic import SupervisorConvenienceConfiguration, SupervisorRemoteXMLRPCClient
from supervisor_pydantic.client.xmlrpc import ProcessState


def _assert_client_actions(client: SupervisorRemoteXMLRPCClient):
    assert client.getProcessInfo("test").state == ProcessState.STOPPED
    sleep(0.5)
    assert client.startAllProcesses()["test"].state == ProcessState.RUNNING
    sleep(0.5)
    assert client.getProcessInfo("test").state == ProcessState.EXITED
    assert client.startProcess("test").state == ProcessState.RUNNING
    assert client.stopProcess("test").state == ProcessState.STOPPED
    assert client.startProcess("test").state == ProcessState.RUNNING
    assert client.stopAllProcesses()["test"].state == ProcessState.STOPPED


def test_supervisor_client(supervisor_instance: SupervisorConvenienceConfiguration):
    client = SupervisorRemoteXMLRPCClient(supervisor_instance)
    _assert_client_actions(client=client)


def test_permissioned_supervisor_client_rejected(permissioned_supervisor_instance: SupervisorConvenienceConfiguration):
    permissioned_supervisor_instance.airflow.username = "bad-username"
    client = SupervisorRemoteXMLRPCClient(permissioned_supervisor_instance)
    with pytest.raises(xmlrpc.client.ProtocolError):
        client.getProcessInfo("test")


def test_permissioned_supervisor_client(permissioned_supervisor_instance: SupervisorConvenienceConfiguration):
    permissioned_supervisor_instance.airflow.username = "user1"
    client = SupervisorRemoteXMLRPCClient(permissioned_supervisor_instance)
    _assert_client_actions(client=client)



from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

from pydantic import Field, field_serializer, field_validator

from .base import OctalUmask, Signal, UnixUserName, _BaseCfgModel

__all__ = ("ProgramConfiguration",)


class ProgramConfiguration(_BaseCfgModel):
    def to_cfg(self, key: str) -> str:
        # Overload to require key
        return super().to_cfg(key=key)

    command: str = Field(
        description="The command that will be run when this program is started. The command can be either absolute (e.g. /path/to/programname) or relative (e.g. programname). If it is relative, the supervisord’s environment $PATH will be searched for the executable. Programs can accept arguments, e.g. /path/to/program foo bar. The command line can use double quotes to group arguments with spaces in them to pass to the program, e.g. /path/to/program/name -p 'foo bar'. Note that the value of command may include Python string expressions, e.g. /path/to/programname --port=80%(process_num)02d might expand to /path/to/programname --port=8000 at runtime. String expressions are evaluated against a dictionary containing the keys group_name, host_node_name, program_name, process_num, numprocs, here (the directory of the supervisord config file), and all supervisord’s environment variables prefixed with ENV_. Controlled programs should themselves not be daemons, as supervisord assumes it is responsible for daemonizing its subprocesses (see Nondaemonizing of Subprocesses). The command will be truncated if it looks like a config file comment, e.g. command=bash -c 'foo ; bar' will be truncated to command=bash -c 'foo. Quoting will not prevent this behavior, since the configuration file reader does not parse the command like a shell would."
    )
    process_name: Optional[str] = Field(
        default=None,
        description="A Python string expression that is used to compose the supervisor process name for this process. You usually don’t need to worry about setting this unless you change numprocs. The string expression is evaluated against a dictionary that includes group_name, host_node_name, process_num, program_name, and here (the directory of the supervisord config file).",
    )
    numprocs: Optional[int] = Field(
        default=None,
        description="Supervisor will start as many instances of this program as named by numprocs. Note that if numprocs > 1, the process_name expression must include %(process_num)s (or any other valid Python string expression that includes process_num) within it.",
    )
    numprocs_start: Optional[int] = Field(
        default=None, description="An integer offset that is used to compute the number at which process_num starts."
    )
    priority: Optional[int] = Field(
        default=None,
        description="The relative priority of the program in the start and shutdown ordering. Lower priorities indicate programs that start first and shut down last at startup and when aggregate commands are used in various clients (e.g. “start all”/”stop all”). Higher priorities indicate programs that start last and shut down first.",
    )
    autostart: Optional[bool] = Field(default=None, description="If true, this program will start automatically when supervisord is started.")
    startsecs: Optional[int] = Field(
        default=None,
        description="The total number of seconds which the program needs to stay running after a startup to consider the start successful (moving the process from the STARTING state to the RUNNING state). Set to 0 to indicate that the program needn’t stay running for any particular amount of time. Even if a process exits with an “expected” exit code (see exitcodes), the start will still be considered a failure if the process exits quicker than startsecs.",
    )
    startretries: Optional[int] = Field(
        default=None,
        description="The number of serial failure attempts that supervisord will allow when attempting to start the program before giving up and putting the process into an FATAL state. After each failed restart, process will be put in BACKOFF state and each retry attempt will take increasingly more time.",
    )
    autorestart: Optional[Union[bool, Literal["unexpected"]]] = Field(
        default=None,
        description="Specifies if supervisord should automatically restart a process if it exits when it is in the RUNNING state. May be one of false, unexpected, or true. If false, the process will not be autorestarted. If unexpected, the process will be restarted when the program exits with an exit code that is not one of the exit codes associated with this process’ configuration (see exitcodes). If true, the process will be unconditionally restarted when it exits, without regard to its exit code. autorestart controls whether supervisord will autorestart a program if it exits after it has successfully started up (the process is in the RUNNING state). supervisord has a different restart mechanism for when the process is starting up (the process is in the STARTING state). Retries during process startup are controlled by startsecs and startretries.",
    )
    exitcodes: Optional[List[int]] = Field(
        default=None,
        description="The list of “expected” exit codes for this program used with autorestart. If the autorestart parameter is set to unexpected, and the process exits in any other way than as a result of a supervisor stop request, supervisord will restart the process if it exits with an exit code that is not defined in this list.",
    )
    stopsignal: Optional[Signal] = Field(
        default=None,
        description="The signal used to kill the program when a stop is requested. This can be specified using the signal’s name or its number. It is normally one of: TERM, HUP, INT, QUIT, KILL, USR1, or USR2.",
    )
    stopwaitsecs: Optional[int] = Field(
        default=None,
        description="The number of seconds to wait for the OS to return a SIGCHLD to supervisord after the program has been sent a stopsignal. If this number of seconds elapses before supervisord receives a SIGCHLD from the process, supervisord will attempt to kill it with a final SIGKILL.",
    )
    stopasgroup: Optional[bool] = Field(
        default=None,
        description="If true, the flag causes supervisor to send the stop signal to the whole process group and implies killasgroup is true. This is useful for programs, such as Flask in debug mode, that do not propagate stop signals to their children, leaving them orphaned.",
    )
    killasgroup: Optional[bool] = Field(
        default=None,
        description="If true, when resorting to send SIGKILL to the program to terminate it send it to its whole process group instead, taking care of its children as well, useful e.g with Python programs using multiprocessing.",
    )
    user: Optional[UnixUserName] = Field(
        default=None,
        description="Instruct supervisord to use this UNIX user account as the account which runs the program. The user can only be switched if supervisord is run as the root user. If supervisord can’t switch to the specified user, the program will not be started. The user will be changed using setuid only. This does not start a login shell and does not change environment variables like USER or HOME. See Subprocess Environment for details.",
    )
    redirect_stderr: Optional[bool] = Field(
        default=None,
        description="If true, cause the process’ stderr output to be sent back to supervisord on its stdout file descriptor (in UNIX shell terms, this is the equivalent of executing /the/program 2>&1). Do not set redirect_stderr=true in an [eventlistener:x] section. Eventlisteners use stdout and stdin to communicate with supervisord. If stderr is redirected, output from stderr will interfere with the eventlistener protocol.",
    )
    stdout_logfile: Optional[Path] = Field(
        default=None,
        description="Put process stdout output in this file (and if redirect_stderr is true, also place stderr output in this file). If stdout_logfile is unset or set to AUTO, supervisor will automatically choose a file location. If this is set to NONE, supervisord will create no log file. AUTO log files and their backups will be deleted when supervisord restarts. The stdout_logfile value can contain Python string expressions that will evaluated against a dictionary that contains the keys group_name, host_node_name, process_num, program_name, and here (the directory of the supervisord config file). It is not possible for two processes to share a single log file (stdout_logfile) when rotation (stdout_logfile_maxbytes) is enabled. This will result in the file being corrupted. If stdout_logfile is set to a special file like /dev/stdout that is not seekable, log rotation must be disabled by setting stdout_logfile_maxbytes = 0.",
    )
    stdout_logfile_maxbytes: Optional[str] = Field(
        default=None,
        description="The maximum number of bytes that may be consumed by stdout_logfile before it is rotated (suffix multipliers like “KB”, “MB”, and “GB” can be used in the value). Set this value to 0 to indicate an unlimited log size.",
    )
    stdout_logfile_backups: Optional[int] = Field(
        default=None,
        description="The number of stdout_logfile backups to keep around resulting from process stdout log file rotation. If set to 0, no backups will be kept.",
    )
    stdout_capture_maxbytes: Optional[int] = Field(
        default=None,
        description="Max number of bytes written to capture FIFO when process is in “stdout capture mode” (see Capture Mode). Should be an integer (suffix multipliers like “KB”, “MB” and “GB” can used in the value). If this value is 0, process capture mode will be off.",
    )
    stdout_events_enabled: Optional[int] = Field(
        default=None,
        description="If true, PROCESS_LOG_STDOUT events will be emitted when the process writes to its stdout file descriptor. The events will only be emitted if the file descriptor is not in capture mode at the time the data is received (see Capture Mode).",
    )
    stdout_syslog: Optional[bool] = Field(default=None, description="If true, stdout will be directed to syslog along with the process name.")
    stderr_logfile: Optional[Path] = Field(
        default=None,
        description="Put process stderr output in this file unless redirect_stderr is true. Accepts the same value types as stdout_logfile and may contain the same Python string expressions. It is not possible for two processes to share a single log file (stderr_logfile) when rotation (stderr_logfile_maxbytes) is enabled. This will result in the file being corrupted. If stderr_logfile is set to a special file like /dev/stderr that is not seekable, log rotation must be disabled by setting stderr_logfile_maxbytes = 0.",
    )
    stderr_logfile_maxbytes: Optional[str] = Field(
        default=None,
        description="The maximum number of bytes before logfile rotation for stderr_logfile. Accepts the same value types as stdout_logfile_maxbytes.",
    )
    stderr_logfile_backups: Optional[int] = Field(
        default=None,
        description="The number of backups to keep around resulting from process stderr log file rotation. If set to 0, no backups will be kept.",
    )
    stderr_capture_maxbytes: Optional[int] = Field(
        default=None,
        description="Max number of bytes written to capture FIFO when process is in “stderr capture mode” (see Capture Mode). Should be an integer (suffix multipliers like “KB”, “MB” and “GB” can used in the value). If this value is 0, process capture mode will be off.",
    )
    stderr_events_enabled: Optional[bool] = Field(
        default=None,
        description="If true, PROCESS_LOG_STDERR events will be emitted when the process writes to its stderr file descriptor. The events will only be emitted if the file descriptor is not in capture mode at the time the data is received (see Capture Mode).",
    )
    stderr_syslog: Optional[bool] = Field(default=None, description="If true, stderr will be directed to syslog along with the process name.")
    environment: Optional[Dict[str, str]] = Field(
        default=None,
        description='A list of key/value pairs in the form KEY="val",KEY2="val2" that will be placed in the child process’ environment. The environment string may contain Python string expressions that will be evaluated against a dictionary containing group_name, host_node_name, process_num, program_name, and here (the directory of the supervisord config file). Values containing non-alphanumeric characters should be quoted (e.g. KEY="val:123",KEY2="val,456"). Otherwise, quoting the values is optional but recommended. Note that the subprocess will inherit the environment variables of the shell used to start “supervisord” except for the ones overridden here. See Subprocess Environment.',
    )
    directory: Optional[Path] = Field(
        default=None,
        description="A file path representing a directory to which supervisord should temporarily chdir before exec’ing the child.",
    )
    umask: Optional[OctalUmask] = Field(default=None, description="An octal number (e.g. 002, 022) representing the umask of the process.")
    serverurl: Optional[str] = Field(
        default=None,
        description="The URL passed in the environment to the subprocess process as SUPERVISOR_SERVER_URL (see supervisor.childutils) to allow the subprocess to easily communicate with the internal HTTP server. If provided, it should have the same syntax and structure as the [supervisorctl] section option of the same name. If this is set to AUTO, or is unset, supervisor will automatically construct a server URL, giving preference to a server that listens on UNIX domain sockets over one that listens on an internet socket.",
    )

    @field_serializer("exitcodes", when_used="json")
    def _dump_exitcodes(self, v):
        if v:
            return ",".join(str(_) for _ in v)
        return None

    @field_validator("exitcodes", mode="before")
    @classmethod
    def _load_exitcodes(cls, v):
        if isinstance(v, str):
            v = v.split(",")
        if isinstance(v, list):
            return [int(_) for _ in v]
        return v

    @field_serializer("autorestart", when_used="json")
    def _dump_autorestart(self, v):
        if isinstance(v, bool):
            return str(v).lower()
        elif isinstance(v, str):
            return v.lower()
        return None

    @field_validator("autorestart", mode="before")
    @classmethod
    def _load_autorestart(cls, v):
        if isinstance(v, str):
            # handle string -> bool
            if v.lower() == "false":
                return False
            if v.lower() == "true":
                return True
        return v
